module cpu(clk, reset,
           bus_en, bus_wr, bus_size, bus_addr,
           bus_data_in, bus_data_out, bus_wt,
           irq);
    input clk;
    input reset;
    output reg bus_en;
    output reg bus_wr;
    output reg [1:0] bus_size;
    output reg [31:0] bus_addr;
    input [31:0] bus_data_in;
    output reg [31:0] bus_data_out;
    input bus_wt;
    input [15:0] irq;

  reg [2:0] state;
  reg [7:0] data;
  reg [6:0] col;
  reg [4:0] row;
  wire [3:0] nybble;
  wire [7:0] hex;
  wire [7:0] dspdat;

  assign nybble =
    (state == 3'd3) ? data[7:4] : data[3:0];
  assign hex =
    (nybble == 4'h0) ? 8'h30 :
    (nybble == 4'h1) ? 8'h31 :
    (nybble == 4'h2) ? 8'h32 :
    (nybble == 4'h3) ? 8'h33 :
    (nybble == 4'h4) ? 8'h34 :
    (nybble == 4'h5) ? 8'h35 :
    (nybble == 4'h6) ? 8'h36 :
    (nybble == 4'h7) ? 8'h37 :
    (nybble == 4'h8) ? 8'h38 :
    (nybble == 4'h9) ? 8'h39 :
    (nybble == 4'ha) ? 8'h41 :
    (nybble == 4'hb) ? 8'h42 :
    (nybble == 4'hc) ? 8'h43 :
    (nybble == 4'hd) ? 8'h44 :
    (nybble == 4'he) ? 8'h45 :
                       8'h46;
  assign dspdat =
    (state == 3'd5) ? 8'h20 : hex;

  always @(posedge clk) begin
    if (reset == 1) begin
      state <= 3'd0;
      col <= 7'd0;
      row <= 5'd0;
      bus_en <= 0;
      bus_wr <= 0;
      bus_size <= 2'b00;
    end else begin
      if (bus_en == 1 && bus_wt == 1) begin
        // wait for bus cycle to complete - do nothing
      end else begin
        case (state)
          3'd0: // read kbd status
                begin
                  state <= 3'd1;
                  bus_addr[1:0] <= 2'b00;       // access by words
                  bus_addr[19:2] <= 18'h0;      // 20 bits per device
                  bus_addr[27:20] <= 8'h02;     // device number
                  bus_addr[31:28] <= 4'h3;      // I/O address space
                  bus_en <= 1;
                  bus_wr <= 0;                  // read
                  bus_size <= 2'b10;            // word
                end
          3'd1: // read kbd data
                begin
                  if (bus_data_in[0] == 1) begin
                    state <= 3'd2;
                    bus_addr[1:0] <= 2'b00;       // access by words
                    bus_addr[19:2] <= 18'h1;      // 20 bits per device
                    bus_addr[27:20] <= 8'h02;     // device number
                    bus_addr[31:28] <= 4'h3;      // I/O address space
                    bus_en <= 1;
                    bus_wr <= 0;                  // read
                    bus_size <= 2'b10;            // word
                  end
                end
          3'd2: // store kbd data
                begin
                  state <= 3'd3;
                  data <= bus_data_in[7:0];
                  bus_en <= 0;
                end
          3'd3: // write high nybble
                begin
                  state <= 3'd4;
                  bus_addr[1:0] <= 2'b00;       // access by words
                  bus_addr[8:2] <= col;         // 20 bits per device
                  bus_addr[13:9] <= row;
                  bus_addr[19:14] <= 6'b000000;
                  bus_addr[27:20] <= 8'h01;     // device number
                  bus_addr[31:28] <= 4'h3;      // I/O address space
                  bus_data_out <= { dspdat, dspdat,
                                    dspdat, dspdat };
                  bus_en <= 1;
                  bus_wr <= 1;                  // write
                  bus_size <= 2'b10;            // word
                  if (col == 7'd77) begin
                    col <= 7'd0;
                    if (row == 5'd29) begin
                      row <= 5'd0;
                    end else begin
                      row <= row + 1;
                    end
                  end else begin
                    col <= col + 1;
                  end
                end
          3'd4: // write low nybble
                begin
                  state <= 3'd5;
                  bus_addr[1:0] <= 2'b00;       // access by words
                  bus_addr[8:2] <= col;         // 20 bits per device
                  bus_addr[13:9] <= row;
                  bus_addr[19:14] <= 6'b000000;
                  bus_addr[27:20] <= 8'h01;     // device number
                  bus_addr[31:28] <= 4'h3;      // I/O address space
                  bus_data_out <= { dspdat, dspdat,
                                    dspdat, dspdat };
                  bus_en <= 1;
                  bus_wr <= 1;                  // write
                  bus_size <= 2'b10;            // word
                  if (col == 7'd77) begin
                    col <= 7'd0;
                    if (row == 5'd29) begin
                      row <= 5'd0;
                    end else begin
                      row <= row + 1;
                    end
                  end else begin
                    col <= col + 1;
                  end
                end
          3'd5: // write blank
                begin
                  state <= 3'd6;
                  bus_addr[1:0] <= 2'b00;       // access by words
                  bus_addr[8:2] <= col;         // 20 bits per device
                  bus_addr[13:9] <= row;
                  bus_addr[19:14] <= 6'b000000;
                  bus_addr[27:20] <= 8'h01;     // device number
                  bus_addr[31:28] <= 4'h3;      // I/O address space
                  bus_data_out <= { dspdat, dspdat,
                                    dspdat, dspdat };
                  bus_en <= 1;
                  bus_wr <= 1;                  // write
                  bus_size <= 2'b10;            // word
                  if (col == 7'd77) begin
                    col <= 7'd0;
                    if (row == 5'd29) begin
                      row <= 5'd0;
                    end else begin
                      row <= row + 1;
                    end
                  end else begin
                    col <= col + 1;
                  end
                end
          3'd6: // end bus cycle
                begin
                  state <= 3'd0;
                  bus_en <= 0;
                  bus_wr <= (| irq[15:0]);      // fake
                  bus_size <= 2'b11;            // fake
                  bus_addr <= bus_data_in;      // fake
                  bus_data_out <= bus_data_in;  // fake
                end
        endcase
      end
    end
  end

endmodule
